local addon = Poppins
if not addon then return end

local LibMacroParser = LibStub("LibMacroParser-1.0")
local LibButtonFacade = LibStub("LibButtonFacade", true)
local LibKeyBound = LibStub("LibKeyBound-1.0")
local media = LibStub("LibSharedMedia-3.0")

local Button = CreateFrame("CheckButton")
LibStub("AceEvent-3.0"):Embed(Button)
Button.callbacks = LibStub("CallbackHandler-1.0"):New(Button)
addon.Button = {}
addon.Button.prototype = Button
addon.Button.metatable = {__index = Button}

function addon.Button:Retrieve(parent)
	local buttonName = ("%sButton"):format(parent:GetName())
	local button = setmetatable(CreateFrame("CheckButton", buttonName, parent, "SecureAnchorEnterTemplate,SecureActionButtonTemplate,ActionButtonTemplate"), addon.Button.metatable)

	button.parent = parent
	button:SetAllPoints(parent)

	button.icon = _G[("%sIcon"):format(buttonName)]
	button.border = _G[("%sBorder"):format(buttonName)]
	button.cooldown = _G[("%sCooldown"):format(buttonName)]
	button.macroName = _G[("%sName"):format(buttonName)]
	button.hotkey = _G[("%sHotKey"):format(buttonName)]
	button.count = _G[("%sCount"):format(buttonName)]
	button.flash = _G[("%sFlash"):format(buttonName)]
	button.flash:Hide()
	
	button:SetAttribute("childstate-onenter", "show"..parent.rid)
	button:SetAttribute("childstate-onleave", "hide"..parent.rid)
	-- attempt to fix modifier-leave not hiding the popbuttons
	button:SetAttribute("*childstate-onenter", "show"..parent.rid)
	button:SetAttribute("*childstate-onleave", "hide"..parent.rid)
	-- end fix
	button:SetAttribute("anchorchild", addon.header)
	parent:SetAttribute("addchild", button)
	parent.button = button

	button.attributes = {}
	button:HookScript("OnAttributeChanged", button.OnAttributeChanged)
	button:HookScript("OnEnter", button.ShowTooltip)
	button:HookScript("OnLeave", button.HideTooltip)
	button:RegisterForClicks("LeftButtonUp", "RightButtonUp")

	button.callbacks:Fire("OnButtonCreate", self)

	return button
end

function Button:Setup()
	LibButtonFacade = LibButtonFacade or LibStub("LibButtonFacade", true)
	if LibButtonFacade then
		LibButtonFacade:Group(addon.name, self.parent.db.group or tostring(self.parent.id)):AddButton(self)
	end

	self.state = ""
	self:RegisterAllEvents()
	self:SetScript("OnShow", self.OnShow)
	self:SetScript("OnHide", self.OnShow)
	self:SetScript("OnDragStart", self.OnDragStart)
	self:SetScript("OnReceiveDrag", self.OnReceiveDrag)
	self:SetScript("PostClick", self.OnClick)
	self.callbacks:Fire("OnButtonSetup", self)
end

function Button:Reset()
	self.callbacks:Fire("OnButtonReset", self)
	LibButtonFacade = LibButtonFacade or LibStub("LibButtonFacade", true)
	if LibButtonFacade then
		LibButtonFacade:Group(addon.name, tostring(self.parent.id)):RemoveButton(self)
	end

	LibMacroParser:UnregisterAllMacros(self)
	self:UnregisterAllEvents()
	self:SetScript("OnShow", nil)
	self:SetScript("OnHide", nil)
	self:SetScript("OnDragStart", nil)
	self:SetScript("OnReceiveDrag", nil)
	self:SetScript("OnUpdate", nil)
	self:SetScript("PostClick", nil)

	for name, value in pairs(self.attributes) do
		self:SetAttribute(name, nil)
	end
end

Button.events = {
	-- Shared
	MODIFIER_STATE_CHANGED = true,
	UPDATE_BINDINGS = true,
	PLAYER_ENTERING_WORLD = true,
	-- Spell
	PLAYER_ENTER_COMBAT = true,
	PLAYER_LEAVE_COMBAT = true,
	SPELL_UPDATE_COOLDOWN = true,
	SPELL_UPDATE_USABLE = true,
	START_AUTOREPEAT_SPELL = true,
	STOP_AUTOREPEAT_SPELL = true,
	PLAYER_TARGET_CHANGED = true,
	UNIT_SPELLCAST_SENT = true,
	UNIT_SPELLCAST_STOP = true,
	UNIT_SPELLCAST_FAILED = true,
	UNIT_SPELLCAST_INTERRUPTED = true,
	CURSOR_UPDATE = true,
	CRAFT_SHOW = true,
	CRAFT_CLOSE = true,
	TRADE_SKILL_SHOW = true,
	TRADE_SKILL_CLOSE = true,
	-- Item
	BAG_UPDATE = true,
	BAG_UPDATE_COOLDOWN = true,
	UNIT_INVENTORY_CHANGED = true,
}

function Button:RegisterAllEvents()
	self:UnregisterAllEvents()
	for event, register in pairs(self.events) do
		if register then
			self:RegisterEvent(event)
		end
	end
end

function Button:OnAttributeChanged(name, value)
	self.attributes[name] = value
end

function Button:OnClick()
	self:UpdateState()
end

function Button:OnShow()
	self:RegisterAllEvents()
	self:Update()
end

function Button:OnHide()
	self:UnregisterAllEvents()
end

local TOOLTIP_UPDATE_TIME = TOOLTIP_UPDATE_TIME
local ATTACK_BUTTON_FLASH_TIME = ATTACK_BUTTON_FLASH_TIME
function Button:OnUpdate(elapsed)
	if self.flashing == 1 then
		self.flashtime = self.flashtime - elapsed
		if self.flashtime <= 0 then
			local overtime = -self.flashtime
			if overtime >= ATTACK_BUTTON_FLASH_TIME then
				overtime = 0
			end
			self.flashtime = ATTACK_BUTTON_FLASH_TIME - overtime
			
			local flashTexture = self.flash
			if flashTexture:IsVisible() then
				flashTexture:Hide()
			else
				flashTexture:Show()
			end
		end
	end

	if self.rangeTimer then
		self.rangeTimer = self.rangeTimer - elapsed
		if self.rangeTimer <= 0 then
			local valid
			local type, value = self.type, self.value
			if type == "spell" then
				valid = IsSpellInRange(value)
			elseif type == "item" then
				valid = IsItemInRange(value)
			end
			self.outOfRange = (valid == 0)
			self:UpdateUsable()
			self.rangeTimer = TOOLTIP_UPDATE_TIME
		end
	end

	if self.bagUpdate then
		self.bagUpdate = self.bagUpdate - elapsed
		if self.bagUpdate <= 0 then
			self:UpdateCount()
			self:UpdateUsable(true)
			self.bagUpdate = nil
		end
	end
end

function Button:ShowTooltip()
	if GetCVar("UberTooltips") == "1" then
		GameTooltip_SetDefaultAnchor(GameTooltip, self)
	else
		GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
	end
	if (self.GetHotkey) then
		LibKeyBound:Set(self)
	end
	local db = addon.db.profile
	if db.showTooltips == 1 then return end
	if db.showTooltips == 3 and InCombatLockdown() then return end
	
	local link, _
	local type, value = self.type, self.value
	if type == "spell" and value then
		link = GetSpellLink(value)
	elseif type == "item" and value then
		_, link = GetItemInfo(value)
	end
	if link and GameTooltip:SetHyperlink(link) then
		self.UpdateTooltip = self.ShowTooltip
	else
		self.UpdateTooltip = nil
	end
end

function Button:HideTooltip()
	GameTooltip:Hide()
end

function Button:MacroUpdate(macro)
	local type, value = self:GetType(true)
	if type == "macro" and value == macro then
		self:Update()
	end
end

function Button:GetType(real)
	local type = SecureButton_GetModifiedAttribute(self, "type"..self.state)
	local value = type and SecureButton_GetModifiedAttribute(self, type..self.state)
	if not real and type == "macro" and value then
		local value = tonumber(value)
		local oldType, oldValue = self.type, self.value
		if type ~= oldType or value ~= oldValue then
			if oldType == "macro" then
				local value = tonumber(oldValue)
				if LibMacroParser:IsMacroRegistered(self, value) then
					LibMacroParser:UnregisterMacro(self, value)
				end
			end
			if not LibMacroParser:IsMacroRegistered(self, value) then
				LibMacroParser:RegisterMacro(self, "MacroUpdate", value)
			end
		end
		return LibMacroParser:GetMacroInfo(value)
	end
	return type, value
end

function Button:Update()
	local type, value = self:GetType()
	self.type, self.value = type, value

	self:UpdateIcon()
	self:UpdateCount()
	self:UpdateHotkeys()
	if value then
		if type == "item" and IsEquippedItem(value) then
			self.border:SetVertexColor(0, 1.0, 0, 0.75)
			self.border:Show()
		else
			self.border:Hide()
		end
		self:UpdateState()
		self:UpdateUsable(true)
		self:UpdateCooldown()
		self:UpdateFlash()
		self:SetScript("OnUpdate", self.OnUpdate)
	else
		self.border:Hide()
		self.cooldown:Hide()
		self:SetScript("OnUpdate", nil)
	end
end

function Button:UpdateIcon()
	local _, texture
	local type, value = self.type, self.value
	if type and value then
		if type == "spell" then
			texture = GetSpellTexture(value)
		elseif type == "item" then
			texture = GetItemIcon(value)
		elseif type == "macro" then
--			before - spit out macro body?!?
--			_, texture = GetMacroInfo(value)
--			now - somewhat bypassing LibMacroparser I guess?
			value = type and SecureButton_GetModifiedAttribute(self, type..self.state)
			value = tonumber(value)
			_, texture = GetMacroInfo(value)
		end
		texture = texture or [[Interface\Icons\INV_Misc_QuestionMark]]
	end

	if texture then
		self.rangeTimer = -1
		self.icon:SetTexture(texture)
		self.icon:Show()
	else
		self.rangeTimer = nil
		self.icon:Hide()
		self.cooldown:Hide()
		self.hotkey:SetVertexColor(0.6, 0.6, 0.6)
	end
end

function Button:UpdateCount()
--	local db = addon.db.profile
	local db = self.parent.db
	local count

	local type, value = self.type, self.value
	if not value then
	elseif type == "spell" and IsConsumableSpell(value) then
		count = GetSpellCount(value)
	elseif type == "item" and IsConsumableItem(value) then
		count = GetItemCount(value, false, true)
	-- temporary solution until theres a macrotext layer. dunno if thats really needed anyways.
	elseif type == "macro" then
			local value
			value = type and SecureButton_GetModifiedAttribute(self, type..self.state)
			value = tonumber(value)
			count = GetMacroInfo(value)
	end
	self.count:SetText(count)
	--LSM
	if media then self.count:SetFont(media:Fetch("font", db.text.count.font), db.text.count.size, "OUTLINE") end
	self.count:SetPoint("TOPLEFT", self, "TOPLEFT", db.text.count.offsetx, db.text.count.offsety)
	if db.text.count.hidden == false then self.count:Show() else self.count:Hide() end
end

function Button:UpdateState()
	local checked = 0
	local type, value = self.type, self.value
	if not value then
	elseif type == "spell" then
		checked = (IsCurrentSpell(value) or IsAutoRepeatSpell(value)) and 1 or 0
	elseif type == "item" then
		checked = IsCurrentItem(value) and 1 or 0
	end
	self:SetChecked(checked)
end

function Button:UpdateUsable(force)
	local isUsable, notEnoughMana
	local icon, hotkey = self.icon, self.hotkey
	if force then 
		icon.state, hotkey.state = 1, 1 
	end

	if self.outOfRange then
		icon:SetVertexColor(0.8, 0.1, 0.1)
	else
		local type, value = self.type, self.value
		if not value then
		elseif type == "spell" then
			isUsable, notEnoughMana = IsUsableSpell(value)
		elseif type == "item" then
			isUsable, notEnoughMana = IsUsableItem(value)
		end

		if isUsable then
			icon:SetVertexColor(1.0, 1.0, 1.0)
		elseif notEnoughMana then
			icon:SetVertexColor(0.5, 0.5, 1.0)
		else
			icon:SetVertexColor(0.4, 0.4, 0.4)
		end
	end
end

function Button:UpdateCooldown()
	local start, duration, enable

	local type, value = self.type, self.value
	if not value then
	elseif type == "spell" then
		start, duration, enable = GetSpellCooldown(value)
	elseif type == "item" then
		start, duration, enable = GetItemCooldown(value)
	end
	start, duration, enable = start or 0, duration or 0, enable or 0

	CooldownFrame_SetTimer(self.cooldown, start, duration, enable)
end

function Button:UpdateFlash()
	local type, value = self.type, self.value
	if type == "spell" and value and (IsAttackSpell(value) and IsCurrentSpell(value)) or IsAutoRepeatSpell(value) then
		self:StartFlash()
	else
		self:StopFlash()
	end
end

function Button:StartFlash()
	local type, value = self.type, self.value
	if type == "spell" and value and IsAttackSpell(value) then
		self.flashing = 1
		self.flashtime = 0
		self:UpdateState()
	end
end

function Button:StopFlash()
	self.flashing = 0
	self.flash:Hide()
	self:UpdateState()
end

function Button:UpdateHotkeys()
--	local db = addon.db.profile
	local db = self.parent.db
	local keybinding = self:GetHotkey() or ""
	keybinding = LibKeyBound:ToShortKey(keybinding)
	local hotkey = self.hotkey

	if keybinding == "" then
		hotkey:SetText(RANGE_INDICATOR)
		hotkey:SetPoint("TOPLEFT", self, "TOPLEFT", 1, -2)
		hotkey:Hide()
	else
		--LSM3
		if media then hotkey:SetFont(media:Fetch("font", db.text.hotkey.font), db.text.hotkey.size, "OUTLINE") end
		hotkey:SetVertexColor(1.0, 1.0, 1.0)
		hotkey:SetText(keybinding)
		hotkey:SetPoint("TOPLEFT", self, "TOPLEFT", db.text.hotkey.offsetx, db.text.hotkey.offsety)
		if db.text.hotkey.hidden == false then hotkey:Show() else hotkey:Hide() end
	end
end
function Button:SetKey(key)
	local frame = self.parent
	frame.db.keybinding = key
	ClearOverrideBindings(self)
	if key then
		SetOverrideBindingClick(self, 1, key, self:GetName())
	end
	self:UpdateHotkeys()
end

function Button:GetHotkey()
	local frame = self.parent
--[[
	local buttonName = ("%sButton"):format(frame:GetName())
	local key1 = GetBindingKey("CLICK "..buttonName..":LeftButton")
	local key = LibKeyBound:ToShortKey(key1)
	return frame.button == self and key
]]--
	return frame.button == self and frame.db.keybinding
end

function Button:UPDATE_BINDINGS()
	self:UpdateHotkeys()
end

function Button:MODIFIER_STATE_CHANGED(modifier, state)
	self:Update()
end

function Button:BAG_UPDATE()
	local type, value = self.type, self.value
	if type == "item" and value and not self.bagUpdate then
		self.bagUpdate = 0.1
	end
end

function Button:BAG_UPDATE_COOLDOWN()
	local type, value = self.type, self.value
	if type == "item" and value then
		self:UpdateCooldown()
	end
end

function Button:UNIT_INVENTORY_CHANGED(event, unit)
	if unit ~= "player" then return end
	local type, value = self.type, self.value
	if type == "item" and value then
		self:Update()
	end
end

function Button:UNIT_SPELLCAST_SENT(event, unit, spell, rank)
	if unit ~= "player" then return end
	self:UpdateState()
end
Button.UNIT_SPELLCAST_STOP = Button.UNIT_SPELLCAST_SENT
Button.UNIT_SPELLCAST_FAILED = Button.UNIT_SPELLCAST_SENT
Button.UNIT_SPELLCAST_INTERRUPTED = Button.UNIT_SPELLCAST_SENT

function Button:PLAYER_ENTER_COMBAT()
	self:StartFlash()
end

function Button:PLAYER_LEAVE_COMBAT()
	self:StopFlash()
end

function Button:SPELL_UPDATE_COOLDOWN()
	local type, value = self.type, self.value
	if type == "spell" and value then
		self:UpdateCooldown()
	end
end

function Button:SPELL_UPDATE_USABLE()
	local type, value = self.type, self.value
	if type == "spell" and value then
		self:UpdateUsable()
	end
end

function Button:START_AUTOREPEAT_SPELL()
	local type, value = self.type, self.value
	if type == "spell" and value and IsAutoRepeatSpell(value) then
		self:StartFlash()
	end
end

function Button:STOP_AUTOREPEAT_SPELL()
	local type, value = self.type, self.value
	if type == "spell" and value and self.flashing == 1 and not IsAttackSpell(value) then
		self:StopFlash()
	end
end

function Button:PLAYER_ENTERING_WORLD()
	self:Update()
end

function Button:PLAYER_TARGET_CHANGED()
	self.rangeTimer = -1
end

function Button:CRAFT_SHOW()
	self:UpdateState()
end
Button.CRAFT_CLOSE = Button.CRAFT_SHOW
Button.TRADE_SKILL_SHOW = Button.CRAFT_SHOW
Button.TRADE_SKILL_CLOSE = Button.CRAFT_SHOW

function Button:CURSOR_UPDATE()
	self:UpdateState()
end
