local addon = Poppins
if not addon then return end

local LibStickyFrames = LibStub("LibStickyFrames-2.0")
local LibKeyBound = LibStub("LibKeyBound-1.0")

local count = 0
local weak = {__mode='k'}
local cache = setmetatable({}, weak)
local used = setmetatable({}, weak)

local Frame = CreateFrame("Frame")
LibStub("AceEvent-3.0"):Embed(Frame)
Frame.callbacks = LibStub("CallbackHandler-1.0"):New(Frame)
addon.Frame = {}
addon.Frame.prototype = Frame
addon.Frame.metatable = {__index = Frame}

function addon.Frame:Retrieve()
	local frame = next(cache)
	if frame then
		cache[frame] = nil
	else
		count = count + 1
		local rid = count
		local header = addon.header
		local frameName = ("%sFrame%d"):format(addon.name, rid)
		frame = setmetatable(CreateFrame("Frame", frameName, header, "SecureStateHeaderTemplate"), addon.Frame.metatable)
		frame.rid = rid
		frame:SetWidth(36)
		frame:SetHeight(36)
		--backdrop for LibKeyBound
		frame:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 16, insets = {left = 0, right = 0, top = 0, bottom = 0},})
		frame:SetBackdropColor(0, 1, 1, 0)	
		local button = addon.Button:Retrieve(frame)
		button.parent = frame
		frame.button = button

		frame.attributes = {}
		frame:HookScript("OnAttributeChanged", frame.OnAttributeChanged)

		frame.popbuttons = {}

		frame.callbacks:Fire("OnFrameCreate", frame, rid)
	end
	used[frame] = true
	return frame
end

function addon.Frame:Iterate()
	return pairs(used)
end

function Frame:Setup(id, db)
	id, db = id or self.id, db or self.db
	self:Reset()
	if not id or not db then return end

	LibButtonFacade = LibButtonFacade or LibStub("LibButtonFacade", true)
	if LibButtonFacade then
		local group = db.group
		if group then
			local db = addon.db.profile.groups[group]
			LibButtonFacade:Group(addon.name, group):Skin(db.skin, db.gloss, db.backdrop, db.colors)
		else
			LibButtonFacade:Group(addon.name, tostring(id)):Skin(db.skin, db.gloss, db.backdrop, db.colors)
		end
	end

	self.id = id
	self.db = db

	LibStickyFrames:RegisterFrame(self)
	LibStickyFrames.RegisterCallback(self, "OnStopFrameMoving")
	LibStickyFrames:SetFrameEnabled(self, true)
	LibStickyFrames:SetFrameInsets(self, -3, -2, -2, -3)
	LibStickyFrames:SetFrameText(self, tostring(id))

	local rid = self.rid
	local header = addon.header
	local button = self.button

	button:Setup()
	--FrameStrata
	button:SetFrameStrata(db.framestrata)
	
	local statebutton = ""
	for i=1, db.count do
		statebutton = ("%s%d:%d;"):format(statebutton, i, i)
	end
	button:SetAttribute("*statebutton*", statebutton)
	local state = statebutton:match(("(%s):%%d-;"):format(db.state)) or "1"
	self:SetAttribute("state", state)

	header:SetAttribute("statemap-anchor-show"..rid, "*:"..id)
	header:SetAttribute("statemap-anchor-hide"..rid, "")
	header:SetAttribute("delaystatemap-anchor-hide"..rid, id..":0")
	header:SetAttribute("delaytimemap-anchor-hide"..rid,  id..":0.5")
	header:SetAttribute("delayhovermap-anchor-hide"..rid, id..":true")

	self:SetScale(db.scale)
	self:SetKeybinding()
	
	for id = 1, db.count do
		self:AddPopButton(id)
	end
	self.callbacks:Fire("OnFrameSetup", self, id, db)
end

function Frame:Reset()
	self.callbacks:Fire("OnFrameReset", self)
	local popbuttons = self.popbuttons
	for id = #popbuttons, 1, -1 do
		popbuttons[id]:Store()
		popbuttons[id] = nil
	end
	LibButtonFacade = LibButtonFacade or LibStub("LibButtonFacade", true)
	if LibButtonFacade then
		LibButtonFacade:DeleteGroup(addon.name, tostring(self.id))
	end

	local rid = self.rid
	local header = addon.header

	LibStickyFrames:UnregisterFrame(self)
	LibStickyFrames.UnregisterAllCallbacks(self)
	for name, value in pairs(self.attributes) do
		self:SetAttribute(name, nil)
	end
	self.button:Reset()

	self:ClearAllPoints()
	self:Hide()

	self.id = nil
	self.db = nil
end

function Frame:Store()
	used[self] = nil

	self:Reset()

	cache[self] = true
end

function Frame:AddPopButton(id)
	local db = self.db
	if id > db.count then return end
	local popButton = addon.PopButton:Retrieve()
	popButton:Setup(id, self, db[id])
	self.popbuttons[id] = popButton
	self:UpdateShowPopButton()
end

function Frame:StateChanged(newState)
	if newState == "nil" then return end
	local oldState = self.db.state
	self.db.state = newState
	local button = self.button
	button.state = "-"..newState
	button:Update()
	self.callbacks:Fire("OnFrameStateChanged", self, newState, oldState)
end

function Frame:SetKeybinding()
	local keybinding = self.db.keybinding
	local button = self.button	
	ClearOverrideBindings(button)
	if keybinding then
		SetOverrideBindingClick(button, 1, keybinding, button:GetName())
	end
	button:UpdateHotkeys()
end

function Frame:LoadPosition()
	local db = self.db
	local anchor = self.anchor
	local pointA, frame, pointB, x, y, s = db.pointA, db.frame, db.pointB, db.x, db.y, self:GetEffectiveScale()

	if frame then
		LibStickyFrames:SetFramePoints(self, pointA, addon.buttons[frame], pointB, x, y)
	else
		x, y = x and x / s, y and y / s
		self:ClearAllPoints()
		self:SetPoint(x and "topleft" or "center", UIParent, y and "bottomleft" or "center", x or 0, y or 0)
		LibStickyFrames:SetFramePoints(self)
	end
	self.callbacks:Fire("OnFrameLoadPosition", self)
end

function Frame:SetLock(v)
	self:SetMovable(not v)
	self.button:EnableMouse(v)

	self.callbacks:Fire("OnFrameSetLock", self, v)
	for id, popbutton in pairs(self.popbuttons) do
		popbutton:SetLock(v)
	end
end

function Frame:OnStopFrameMoving(event, frame, pointA, frameB, pointB, x, y)
	if frame ~= self then return end
	local db = self.db
	db.frame = false
	db.pointA = false
	db.pointB = false
	
	if frameB and (frameB:GetParent() == addon.header) and not frameB.parent then
		db.frame = frameB.id
		db.pointA = pointA
		db.pointB = pointB
		db.x = x
		db.y = y
	else
		local x, y, s = self:GetLeft(), self:GetTop(), self:GetEffectiveScale()
		x, y = x * s, y  * s
	
		db.x = x
		db.y = y
	end
	self:LoadPosition()
end

function Frame:OnAttributeChanged(name, value)
	self.attributes[name] = value
end

--LibKeyBound related functions
local colorMoveButtons = {r = 1, b = 1, g = 0, a = 0.5}

function Frame:ColorBars()
	if (Poppins.assignBindings) or (Poppins.stickyMode) then --or AutoBar.unlockButtons) then
		-- Adjust Frame Strata
		self:SetFrameStrata("DIALOG")
		self.button:SetFrameStrata("LOW")
		-- Set Color
		if (Poppins.assignBindings) then
			self:SetBackdropColor(LibKeyBound:GetColorKeyBoundMode())
		elseif (Poppins.stickyMode) then
			self:SetBackdropColor(colorMoveButtons.r, colorMoveButtons.g, colorMoveButtons.b, colorMoveButtons.a)
		end
	else
		self:SetFrameStrata("BACKGROUND")
		self.button:SetFrameStrata(self.db.framestrata)
		self:SetBackdropColor(0, 0, 0, 0)
		self:SetBackdropBorderColor(0, 0, 0, 0)
	end
end

function Frame:UpdateShowPopButton()
	-- if Frame has only one popbutton AND it has contents AND the buttons are locked dont show the PopButton on mouseover
	-- called by Frame:AddPopButton, addon.db.profile.lock (set), 
	local type = self.popbuttons[1].type
	if self.db.count == 1 and type and addon.db.profile.lock then
		self.button:SetAttribute("childstate-onenter", nil)
		self.button:SetAttribute("*childstate-onenter", nil)
	else
		self.button:SetAttribute("childstate-onenter", "show"..self.rid)
		self.button:SetAttribute("*childstate-onenter", "show"..self.rid)
	end
end
